;; Generated package description from treesit-fold.el  -*- mode: lisp-data; no-byte-compile: t; lexical-binding:t -*-
(define-package "treesit-fold" "0.2.1.0.20260226.70748" "Code folding using treesit" '((emacs "29.1")) :commit "d70c5f7240a8a48819421260c9a018c884a41111" :authors '(("Junyi Hou" . "junyi.yi.hou@gmail.com") ("Jen-Chieh" . "jcs090218@gmail.com")) :maintainer '("Jen-Chieh" . "jcs090218@gmail.com") :keywords '("convenience" "folding" "tree-sitter") :url "https://github.com/emacs-tree-sitter/treesit-fold")
